@@ -1,4 +1,13 @@
-//! Securely zero memory with a simple trait ([Zeroize]) built on stable Rust
+#![no_std]
+#![cfg_attr(docsrs, feature(doc_cfg))]
+#![doc(
+    html_logo_url = "https://raw.githubusercontent.com/RustCrypto/media/6ee8e381/logo.svg",
+    html_favicon_url = "https://raw.githubusercontent.com/RustCrypto/media/6ee8e381/logo.svg",
+    html_root_url = "https://docs.rs/zeroize/1.5.1"
+)]
+#![warn(missing_docs, rust_2018_idioms, unused_qualifications)]
+
+//! Securely zero memory with a simple trait ([`Zeroize`]) built on stable Rust
 //! primitives which guarantee the operation will not be "optimized away".
 //!
 //! ## About
@@ -8,8 +17,8 @@
 //! many documented "tricks" to attempt to avoid these optimizations and ensure
 //! that a zeroing routine is performed reliably.
 //!
-//! This crate isn't about tricks: it uses [core::ptr::write_volatile]
-//! and [core::sync::atomic] memory fences to provide easy-to-use, portable
+//! This crate isn't about tricks: it uses [`core::ptr::write_volatile`]
+//! and [`core::sync::atomic`] memory fences to provide easy-to-use, portable
 //! zeroing behavior which works on all of Rust's core number types and slices
 //! thereof, implemented in pure Rust with no usage of FFI or assembly.
 //!
@@ -44,19 +53,19 @@
 //! }
 //! ```
 //!
-//! The [Zeroize] trait is impl'd on all of Rust's core scalar types including
+//! The [`Zeroize`] trait is impl'd on all of Rust's core scalar types including
 //! integers, floats, `bool`, and `char`.
 //!
 //! Additionally, it's implemented on slices and `IterMut`s of the above types.
 //!
 //! When the `alloc` feature is enabled (which it is by default), it's also
 //! impl'd for `Vec<T>` for the above types as well as `String`, where it provides
-//! [Vec::clear()] / [String::clear()]-like behavior (truncating to zero-length)
+//! [`Vec::clear`] / [`String::clear`]-like behavior (truncating to zero-length)
 //! but ensures the backing memory is securely zeroed with some caveats.
 //! (NOTE: see "Stack/Heap Zeroing Notes" for important `Vec`/`String` details)
 //!
-//! The [DefaultIsZeroes] marker trait can be impl'd on types which also
-//! impl [Default], which implements [Zeroize] by overwriting a value with
+//! The [`DefaultIsZeroes`] marker trait can be impl'd on types which also
+//! impl [`Default`], which implements [`Zeroize`] by overwriting a value with
 //! the default value.
 //!
 //! ## Custom Derive Support
@@ -66,20 +75,30 @@
 //! which automatically calls `zeroize()` on all members of a struct
 //! or tuple struct.
 //!
-//! Additionally it supports the following attribute:
+//! Attributes supported for `Zeroize`:
 //!
-//! - `#[zeroize(drop)]`: call `zeroize()` when this item is dropped
+//! On the item level:
+//! - `#[zeroize(drop)]`: *deprecated* use `ZeroizeOnDrop` instead
+//! - `#[zeroize(bound = "T: MyTrait")]`: this replaces any trait bounds
+//!   inferred by zeroize
+//!
+//! On the field level:
+//! - `#[zeroize(skip)]`: skips this field or variant when calling `zeroize()`
+//!
+//! Attributes supported for `ZeroizeOnDrop`:
+//!
+//! On the field level:
+//! - `#[zeroize(skip)]`: skips this field or variant when calling `zeroize()`
 //!
 //! Example which derives `Drop`:
 //!
 //! ```
-//! # #[cfg(feature = "derive")]
+//! # #[cfg(feature = "zeroize_derive")]
 //! # {
-//! use zeroize::Zeroize;
+//! use zeroize::{Zeroize, ZeroizeOnDrop};
 //!
 //! // This struct will be zeroized on drop
-//! #[derive(Zeroize)]
-//! #[zeroize(drop)]
+//! #[derive(Zeroize, ZeroizeOnDrop)]
 //! struct MyStruct([u8; 32]);
 //! # }
 //! ```
@@ -87,7 +106,7 @@
 //! Example which does not derive `Drop` (useful for e.g. `Copy` types)
 //!
 //! ```
-//! #[cfg(feature = "derive")]
+//! #[cfg(feature = "zeroize_derive")]
 //! # {
 //! use zeroize::Zeroize;
 //!
@@ -97,6 +116,19 @@
 //! # }
 //! ```
 //!
+//! Example which only derives `Drop`:
+//!
+//! ```
+//! # #[cfg(feature = "zeroize_derive")]
+//! # {
+//! use zeroize::ZeroizeOnDrop;
+//!
+//! // This struct will be zeroized on drop
+//! #[derive(ZeroizeOnDrop)]
+//! struct MyStruct([u8; 32]);
+//! # }
+//! ```
+//!
 //! ## `Zeroizing<Z>`: wrapper for zeroizing arbitrary values on drop
 //!
 //! `Zeroizing<Z: Zeroize>` is a generic wrapper type that impls `Deref`
@@ -135,10 +167,11 @@
 //! [these remarks have been removed] and the specific usage pattern in this
 //! crate is considered to be well-defined.
 //!
-//! Additionally this crate leverages [compiler_fence] from
-//! [core::sync::atomic] with the strictest ordering ([Ordering::SeqCst])
-//! as a precaution to help ensure reads are not reordered before memory has
-//! been zeroed.
+//! Additionally this crate leverages [`core::sync::atomic::compiler_fence`]
+//! with the strictest ordering
+//! ([`Ordering::SeqCst`]) as a
+//! precaution to help ensure reads are not reordered before memory has been
+//! zeroed.
 //!
 //! All of that said, there is still potential for microarchitectural attacks
 //! (ala Spectre/Meltdown) to leak "zeroized" secrets through covert channels.
@@ -152,12 +185,12 @@
 //! However, be aware several operations in Rust can unintentionally leave
 //! copies of data in memory. This includes but is not limited to:
 //!
-//! - Moves and `Copy`
-//! - Heap reallocation when using `Vec` and `String`
+//! - Moves and [`Copy`]
+//! - Heap reallocation when using [`Vec`] and [`String`]
 //! - Borrowers of a reference making copies of the data
 //!
-//! [`Pin`][pin] can be leveraged in conjunction with this crate to ensure
-//! data kept on the stack isn't moved.
+//! [`Pin`][`core::pin::Pin`] can be leveraged in conjunction with this crate
+//! to ensure data kept on the stack isn't moved.
 //!
 //! The `Zeroize` impls for `Vec` and `String` zeroize the entire capacity of
 //! their backing buffer, but cannot guarantee copies of the data were not
@@ -191,25 +224,11 @@
 //! `unsafe` memory protection systems and just trying to make the best memory
 //! zeroing crate available.
 //!
-//! [Zeroize]: https://docs.rs/zeroize/latest/zeroize/trait.Zeroize.html
 //! [Zeroing memory securely is hard]: http://www.daemonology.net/blog/2014-09-04-how-to-zero-a-buffer.html
-//! [Vec::clear()]: https://doc.rust-lang.org/std/vec/struct.Vec.html#method.clear
-//! [String::clear()]: https://doc.rust-lang.org/std/string/struct.String.html#method.clear
-//! [DefaultIsZeroes]: https://docs.rs/zeroize/latest/zeroize/trait.DefaultIsZeroes.html
-//! [Default]: https://doc.rust-lang.org/std/default/trait.Default.html
-//! [core::ptr::write_volatile]: https://doc.rust-lang.org/core/ptr/fn.write_volatile.html
 //! [Unsafe Code Guidelines Working Group]: https://github.com/rust-lang/unsafe-code-guidelines
 //! [these remarks have been removed]: https://github.com/rust-lang/rust/pull/60972
-//! [core::sync::atomic]: https://doc.rust-lang.org/stable/core/sync/atomic/index.html
-//! [Ordering::SeqCst]: https://doc.rust-lang.org/std/sync/atomic/enum.Ordering.html#variant.SeqCst
-//! [compiler_fence]: https://doc.rust-lang.org/stable/core/sync/atomic/fn.compiler_fence.html
-//! [pin]: https://doc.rust-lang.org/std/pin/struct.Pin.html
 //! [good cryptographic hygiene]: https://github.com/veorq/cryptocoding#clean-memory-of-secret-data
-
-#![no_std]
-#![cfg_attr(docsrs, feature(doc_cfg))]
-#![doc(html_root_url = "https://docs.rs/zeroize/1.4.1")]
-#![warn(missing_docs, rust_2018_idioms, unused_qualifications)]
+//! [`Ordering::SeqCst`]: core::sync::atomic::Ordering::SeqCst
 
 #[cfg(feature = "alloc")]
 #[cfg_attr(test, macro_use)]
@@ -217,12 +236,17 @@ extern crate alloc;
 
 #[cfg(feature = "zeroize_derive")]
 #[cfg_attr(docsrs, doc(cfg(feature = "zeroize_derive")))]
-pub use zeroize_derive::Zeroize;
+pub use zeroize_derive::{Zeroize, ZeroizeOnDrop};
 
 #[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
 mod x86;
 
+use core::marker::{PhantomData, PhantomPinned};
 use core::mem::{self, MaybeUninit};
+use core::num::{
+    NonZeroI128, NonZeroI16, NonZeroI32, NonZeroI64, NonZeroI8, NonZeroIsize, NonZeroU128,
+    NonZeroU16, NonZeroU32, NonZeroU64, NonZeroU8, NonZeroUsize,
+};
 use core::{ops, ptr, slice::IterMut, sync::atomic};
 
 #[cfg(feature = "alloc")]
@@ -235,6 +259,34 @@ pub trait Zeroize {
     fn zeroize(&mut self);
 }
 
+/// Marker trait signifying that this type will [`zeroize`](Zeroize::zeroize) itself on [`Drop`].
+pub trait ZeroizeOnDrop {}
+
+#[doc(hidden)]
+pub mod __internal {
+    use super::*;
+
+    /// Auto-deref workaround for deriving `ZeroizeOnDrop`.
+    pub trait AssertZeroizeOnDrop {
+        fn zeroize_or_on_drop(self);
+    }
+
+    impl<T: ZeroizeOnDrop> AssertZeroizeOnDrop for &mut T {
+        fn zeroize_or_on_drop(self) {}
+    }
+
+    /// Auto-deref workaround for deriving `ZeroizeOnDrop`.
+    pub trait AssertZeroize {
+        fn zeroize_or_on_drop(&mut self);
+    }
+
+    impl<T: Zeroize> AssertZeroize for T {
+        fn zeroize_or_on_drop(&mut self) {
+            self.zeroize()
+        }
+    }
+}
+
 /// Marker trait for types whose `Default` is the desired zeroization result
 pub trait DefaultIsZeroes: Copy + Default + Sized {}
 
@@ -258,6 +310,35 @@ impl_zeroize_with_default!(i8, i16, i32, i64, i128, isize);
 impl_zeroize_with_default!(u8, u16, u32, u64, u128, usize);
 impl_zeroize_with_default!(f32, f64, char, bool);
 
+macro_rules! impl_zeroize_for_non_zero {
+    ($($type:ty),+) => {
+        $(impl Zeroize for $type
+        {
+            fn zeroize(&mut self) {
+                volatile_write(self, unsafe { <$type>::new_unchecked(1) });
+                atomic_fence();
+            }
+        })+
+    };
+}
+
+impl_zeroize_for_non_zero!(
+    NonZeroI8,
+    NonZeroI16,
+    NonZeroI32,
+    NonZeroI64,
+    NonZeroI128,
+    NonZeroIsize
+);
+impl_zeroize_for_non_zero!(
+    NonZeroU8,
+    NonZeroU16,
+    NonZeroU32,
+    NonZeroU64,
+    NonZeroU128,
+    NonZeroUsize
+);
+
 /// Implement `Zeroize` on arrays of types that impl `Zeroize`
 impl<Z, const N: usize> Zeroize for [Z; N]
 where
@@ -267,6 +348,8 @@ where
         self.iter_mut().zeroize();
     }
 }
+/// Implement `ZeroizeOnDrop` on arrays of types that impl `ZeroizeOnDrop`
+impl<Z, const N: usize> ZeroizeOnDrop for [Z; N] where Z: ZeroizeOnDrop {}
 
 impl<'a, Z> Zeroize for IterMut<'a, Z>
 where
@@ -319,6 +402,8 @@ where
     }
 }
 
+impl<Z> ZeroizeOnDrop for Option<Z> where Z: ZeroizeOnDrop {}
+
 /// Impl `Zeroize` on slices of MaybeUninit types
 /// This impl can eventually be optimized using an memset intrinsic,
 /// such as `core::intrinsics::volatile_set_memory`.
@@ -396,6 +481,10 @@ where
     }
 }
 
+#[cfg(feature = "alloc")]
+#[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
+impl<Z> ZeroizeOnDrop for Vec<Z> where Z: ZeroizeOnDrop {}
+
 #[cfg(feature = "alloc")]
 #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
 impl<Z> Zeroize for Box<[Z]>
@@ -409,6 +498,10 @@ where
     }
 }
 
+#[cfg(feature = "alloc")]
+#[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
+impl<Z> ZeroizeOnDrop for Box<[Z]> where Z: ZeroizeOnDrop {}
+
 #[cfg(feature = "alloc")]
 #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
 impl Zeroize for String {
@@ -432,7 +525,7 @@ pub trait TryZeroize {
 
 /// `Zeroizing` is a a wrapper for any `Z: Zeroize` type which implements a
 /// `Drop` handler which zeroizes dropped values.
-#[derive(Debug, Eq, PartialEq)]
+#[derive(Debug, Default, Eq, PartialEq)]
 pub struct Zeroizing<Z: Zeroize>(Z);
 
 impl<Z> Zeroizing<Z>
@@ -495,6 +588,8 @@ where
     }
 }
 
+impl<Z> ZeroizeOnDrop for Zeroizing<Z> where Z: Zeroize {}
+
 impl<Z> Drop for Zeroizing<Z>
 where
     Z: Zeroize,
@@ -544,12 +639,103 @@ unsafe fn volatile_set<T: Copy + Sized>(dst: *mut T, src: T, count: usize) {
     }
 }
 
+/// `PhantomData` is always zero sized so provide a Zeroize implementation.
+impl<Z> Zeroize for PhantomData<Z> {
+    fn zeroize(&mut self) {}
+}
+/// `PhantomData` is always zero sized so provide a ZeroizeOnDrop implementation.
+impl<Z> ZeroizeOnDrop for PhantomData<Z> {}
+/// `PhantomPinned` is zero sized so provide a Zeroize implementation.
+impl Zeroize for PhantomPinned {
+    fn zeroize(&mut self) {}
+}
+/// `PhantomPinned` is zero sized so provide a ZeroizeOnDrop implementation.
+impl ZeroizeOnDrop for PhantomPinned {}
+/// `()` is zero sized so provide a Zeroize implementation.
+impl Zeroize for () {
+    fn zeroize(&mut self) {}
+}
+/// `()` is zero sized so provide a ZeroizeOnDrop implementation.
+impl ZeroizeOnDrop for () {}
+
+/// Generic implementation of Zeroize for tuples up to 10 parameters.
+impl<A: Zeroize> Zeroize for (A,) {
+    fn zeroize(&mut self) {
+        self.0.zeroize();
+    }
+}
+/// Generic implementation of ZeroizeOnDrop for tuples up to 10 parameters.
+impl<A: ZeroizeOnDrop> ZeroizeOnDrop for (A,) {}
+macro_rules! impl_zeroize_tuple {
+    ( $( $type_name:ident )+ ) => {
+        impl<$($type_name: Zeroize),+> Zeroize for ($($type_name),+) {
+            fn zeroize(&mut self) {
+                #[allow(non_snake_case)]
+                let ($($type_name),+) = self;
+                $($type_name.zeroize());+
+            }
+        }
+
+        impl<$($type_name: ZeroizeOnDrop),+> ZeroizeOnDrop for ($($type_name),+) { }
+    }
+}
+// Generic implementations for tuples up to 10 parameters.
+impl_zeroize_tuple! { A B }
+impl_zeroize_tuple! { A B C }
+impl_zeroize_tuple! { A B C D }
+impl_zeroize_tuple! { A B C D E }
+impl_zeroize_tuple! { A B C D E F }
+impl_zeroize_tuple! { A B C D E F G }
+impl_zeroize_tuple! { A B C D E F G H }
+impl_zeroize_tuple! { A B C D E F G H I }
+impl_zeroize_tuple! { A B C D E F G H I L }
+
 #[cfg(test)]
 mod tests {
     use super::*;
+
+    use core::mem::size_of;
+
     #[cfg(feature = "alloc")]
     use alloc::boxed::Box;
 
+    #[derive(Clone, Debug, PartialEq)]
+    struct ZeroizedOnDrop(u64);
+
+    impl Drop for ZeroizedOnDrop {
+        fn drop(&mut self) {
+            self.0.zeroize();
+        }
+    }
+
+    #[test]
+    fn non_zero() {
+        macro_rules! non_zero_test {
+            ($($type:ty),+) => {
+                $(let mut value = <$type>::new(42).unwrap();
+                value.zeroize();
+                assert_eq!(value.get(), 1);)+
+            };
+        }
+
+        non_zero_test!(
+            NonZeroI8,
+            NonZeroI16,
+            NonZeroI32,
+            NonZeroI64,
+            NonZeroI128,
+            NonZeroIsize
+        );
+        non_zero_test!(
+            NonZeroU8,
+            NonZeroU16,
+            NonZeroU32,
+            NonZeroU64,
+            NonZeroU128,
+            NonZeroUsize
+        );
+    }
+
     #[test]
     fn zeroize_byte_arrays() {
         let mut arr = [42u8; 137];
@@ -557,6 +743,13 @@ mod tests {
         assert_eq!(arr.as_ref(), [0u8; 137].as_ref());
     }
 
+    #[test]
+    fn zeroize_on_drop_byte_arrays() {
+        let mut arr = [ZeroizedOnDrop(42); 1];
+        unsafe { core::ptr::drop_in_place(&mut arr) };
+        assert_eq!(arr.as_ref(), [ZeroizedOnDrop(0); 1].as_ref());
+    }
+
     #[test]
     fn zeroize_maybeuninit_byte_arrays() {
         let mut arr = [MaybeUninit::new(42u64); 64];
@@ -565,6 +758,37 @@ mod tests {
         assert_eq!(arr_init, [0u64; 64]);
     }
 
+    #[test]
+    fn zeroize_check_zerosize_types() {
+        // Since we assume these types have zero size, we test this holds for
+        // the current version of Rust.
+        assert_eq!(size_of::<()>(), 0);
+        assert_eq!(size_of::<PhantomPinned>(), 0);
+        assert_eq!(size_of::<PhantomData<usize>>(), 0);
+    }
+
+    #[test]
+    fn zeroize_check_tuple() {
+        let mut tup1 = (42u8,);
+        tup1.zeroize();
+        assert_eq!(tup1, (0u8,));
+
+        let mut tup2 = (42u8, 42u8);
+        tup2.zeroize();
+        assert_eq!(tup2, (0u8, 0u8));
+    }
+
+    #[test]
+    fn zeroize_on_drop_check_tuple() {
+        let mut tup1 = (ZeroizedOnDrop(42),);
+        unsafe { core::ptr::drop_in_place(&mut tup1) };
+        assert_eq!(tup1, (ZeroizedOnDrop(0),));
+
+        let mut tup2 = (ZeroizedOnDrop(42), ZeroizedOnDrop(42));
+        unsafe { core::ptr::drop_in_place(&mut tup2) };
+        assert_eq!(tup2, (ZeroizedOnDrop(0), ZeroizedOnDrop(0)));
+    }
+
     #[cfg(feature = "alloc")]
     #[test]
     fn zeroize_vec() {
